/*
 * Registry RSSI 
 *
 * Author: Liu Rui
 * 
 * RSS (Regular Statement String) is a development tool for software
 * application venders or developers to build their friendly user interfaces or
 * represent their informations.
 *  
 * You may use this software freely and may ask for help from Liu Rui as you 
 * meet any technical problems about it. But Liu Rui does not guarantee your 
 * applications to be well done.
 */

/******************************************************************************/
/*---------------------------- HEADER FILES ----------------------------------*/
#include "registry.h"


unsigned IniFileLineMaxLen=10000;

#if defined(_WIN32) || defined(WIN32) || defined(WIN16)
 char TxtLnEd[]= "\r\n";
#else
 char TxtLnEd[]= "\n";
#endif


/******************************************************************************/
/*Functions*/
/******************************************************************************/
#ifndef WIN32
#include <unistd.h>

/******************************************************************************/


/*
 */
static int FileInsStr(FILE *inif, long fl, char *ins_buf){
	char buf[1024];
	int ok=-1, rn, r, len;
	long l;


	len= strlen(ins_buf);
	fseek(inif, 0, 2);
	l= ftell(inif);

for( rn=1024; l>fl && rn==1024; l-=rn ){
	if(l>= fl+1024)
	rn=1024;
	else
	rn= l-fl;

	fseek(inif, l-rn, 0);
	fread(buf, 1, rn, inif);
	fseek(inif, l-rn+len, 0);
	fwrite(buf, 1, rn, inif);
}	

	r= fseek(inif, fl, 0);
	if(r) goto EX;
	fwrite(ins_buf, 1, len, inif);

	ok=0;
EX:
	return ok;
}


/*
 */
static int FileCut0(char *inifs, long fl, int len){
	char buf[1024];
	int ok=-1, rn, fn;
	long flen, l;
	FILE *inif=0;

	if(len<=0) goto EX;

	inif= fopen( inifs, "r+b" );
	if(!inif) goto EX;

	fseek(inif, 0, 2);
	flen= ftell(inif);

	l= fl+len;
	if(l>=flen) goto CUT;

for( rn=1024; rn==1024; l+=rn ){
	fseek(inif, l, 0);
	rn= fread(buf, 1, 1024, inif);

	fseek(inif, l-len, 0);
	fwrite(buf, 1, rn, inif);
}

CUT:
#if defined(WIN32)
	fn= _fileno(inif);
	_chsize(fn, flen-len);
#else
	fn= fileno(inif);
	ftruncate(fn, flen-len);
#endif

	ok=0;
EX:
	if(inif)
	fclose(inif);

	return ok;
}


/*
 */
static int FileGetLineLen(FILE *inif, long fl){
	char *buf=0;
	int len=0;

	buf= (char *)malloc(IniFileLineMaxLen +1);
	if(!buf) goto EX;

	fseek(inif, fl, 0);
	if( !fgets(buf, IniFileLineMaxLen, inif) ) goto EX;

	len= strlen(buf);

EX:
	if(buf)
	free(buf);

	return len;
}


/*
 */
static int FileNextLine(FILE *inif, long *pfl){
	char *buf=0;
	int ok=-1;

	buf= (char *)malloc(IniFileLineMaxLen +1);
	if(!buf) goto EX;

	fseek(inif, *pfl, 0);
	if( !fgets(buf, IniFileLineMaxLen, inif) ) goto EX;
	*pfl= ftell(inif);

	ok=0;
EX:
	if(buf)
	free(buf);

	return ok;
}


/*************************************************************************/

/*
 */
static int IniFindSect( FILE *inif, char *sect, long *pfl ){
	int ok=-1, sect_n;
	char *buf=0;

	buf= (char *)malloc(IniFileLineMaxLen +1);
	if(!buf) goto EX;

	sect_n= strlen(sect);
	fseek(inif, *pfl, 0);
while( fgets(buf, IniFileLineMaxLen, inif) ){
 if( ( '[' == *buf ) && !memcmp( sect, buf+1, sect_n ) && ( ']' == buf[1+sect_n] ) ){
	ok=0;
	break;
 }

	*pfl= ftell(inif);
}

EX:
	if(buf)
	free(buf);

	return ok;
}


/*
 */
static int Ini_PeekSect( FILE *inif, char *sect, long len, long *pfl ){
	int ok=-1, n;
	char *buf=0, *p;

	buf= (char *)malloc(IniFileLineMaxLen +1);
	if(!buf) goto EX;

	fseek(inif, *pfl, 0);
while( fgets(buf, IniFileLineMaxLen, inif) ){
 if( '[' == *buf ){
	p= strchr( buf, ']' );
	if(!p)continue;

	n= p-buf-1;
	if(n> len-1)
	n= len-1;
	memcpy( sect, buf+1, n );
	sect[n] = '\0';

	*pfl= ftell(inif);

	ok=0;
	break;
 }

	*pfl= ftell(inif);
}

EX:
	if(buf)
	free(buf);

	return ok;
}


/*
 */
static int IniFindEntry( FILE *inif, char *entry, long *pfl ){
	int ok=-1, entry_n;
	char *buf=0;

	buf= (char *)malloc(IniFileLineMaxLen +1);
	if(!buf) goto EX;

	entry_n= strlen(entry);
	fseek(inif, *pfl, 0);
while( fgets(buf, IniFileLineMaxLen, inif) ){
 if( !memcmp( entry, buf, entry_n ) && ( '=' == buf[entry_n] ) ){
	ok=0;
	break;
 }
	if( '[' == *buf ) break;

	*pfl= ftell(inif);
}

EX:
	if(buf)
	free(buf);

	return ok;
}


/*
 */
static int IniGetValue( FILE *inif, long fl, char *vlu, long len ){
	int ok=-1;
	char *buf=0, *p;

	if(len<=0) goto EX;

	buf= (char *)malloc(IniFileLineMaxLen +1);
	if(!buf) goto EX;

	fseek(inif, fl, 0);
	if( !fgets(buf, IniFileLineMaxLen, inif) ) goto EX;

	p= strchr( buf, '=' );
	if(!p) goto EX;

	strncpy( vlu, p+1, len );
	vlu[len-1] = '\0';

	for( p= vlu+len-1; (p>=vlu) && (*p <= ' '); p-- )
	*p = '\0';

	ok=0;
EX:
	if(buf)
	free(buf);

	return ok;
}


/*************************************************************************/

/*
 */
static int IniPeekSect(char *sect, long len, long *pfl, char *inifs){
	int ok=-1, r;
	FILE *inif=0;


	inif= fopen( inifs, "rb" );
	if(!inif) goto EX;

/* */
	ok=1;

	r= Ini_PeekSect(inif, sect, len, pfl);
	if(r<0) goto EX;		/* End of file */

	ok=0;
EX:
	if(inif)
	fclose(inif);

	return ok;
}


/*
 */
static int IniGetLine(char *sect, char *entry, char *dflt, char *vlu, long len, char *inifs){
	int ok=-1, r;
	long fl;
	FILE *inif=0;


	strncpy(vlu, dflt, len);

	inif= fopen( inifs, "rb" );
	if(!inif) goto EX;

/* */
	ok=1;

	fl=0;
	r= IniFindSect(inif, sect, &fl);
	if(r<0) goto EX;		/* End of file */

	r= FileNextLine(inif, &fl);
	if(r<0) goto EX;

	r= IniFindEntry(inif, entry, &fl);
	if(r<0) goto EX;		/* End of file or Next section */

	r= IniGetValue(inif, fl, vlu, len);
	if(r<0) goto EX;

	ok=0;
EX:
	if(inif)
	fclose(inif);

	*(vlu +len -1)=0;
	return ok;
}


/*
 */
static int IniSetLine(char *sect,char *entry,char *str,char *inifs){
	char *ins_buf=0;
	int ok=-1, r, len;
	long fl, fl0;
	FILE *inif=0;


	inif= fopen( inifs, "ab" );
	if(!inif) goto EX;
	fclose(inif);

	inif= fopen( inifs, "r+b" );
	if(!inif) goto EX;

/* */
	ok=1;

	fl=0;
	r= IniFindSect(inif, sect, &fl);
if(r<0){ /* End of file */
 if(str){
	fseek(inif, 0, 2);
	fprintf( inif, "%s[%s]%s%s=%s%s", TxtLnEd, sect, TxtLnEd, entry, str, TxtLnEd );
 }
	goto EX;
}

	r= FileNextLine(inif, &fl);
	if(r<0) goto EX;

	r= IniFindEntry(inif, entry, &fl);
if(r>=0){
	fl0=fl;

	len= FileGetLineLen(inif, fl);

	fclose(inif);
	inif=0;

	r= FileCut0(inifs, fl, len);
	if(r<0) goto EX;

	inif= fopen( inifs, "r+b" );
	if(!inif) goto EX;

	fl=fl0;
}

if(str){
	ins_buf= (char *)malloc(IniFileLineMaxLen +1);
	sprintf( ins_buf, "%s=%s%s", entry, str, TxtLnEd );
	FileInsStr(inif, fl, ins_buf);
	free(ins_buf);
}

	ok =0;
EX:
	if(inif)
	fclose(inif);

	return ok;
}


/*************************************************************************/
/* !WIN32 */
#endif
/*************************************************************************/


/*
 */
RSSI_DLL_EXPORT void RegistryRssi(Rsst *rsst, char *act){
	int rt, status;
	char obj[512], s[8192], s1[512];
	Rsst *h_rsst;
#ifdef WIN32
	HKEY hkey;
	DWORD datatype, datalen;
#endif


if( !strcmp( act, "Activate" ) ){
	h_rsst= RsstNewHandle(rsst);

	RsstKeyCopy( rsst, h_rsst, "RegName" );

	RsstClear(rsst);
	RsstSetStatus( rsst, RSS_STATUS_START );
}
else if( !strcmp( act, "Deactivate" ) ){
	RsstCloseHandle(rsst);
}
else if( !strcmp( act, "ReadKey" ) ){
	h_rsst= RsstGetHandle(rsst);
	status= RsstGetStatus(rsst);
 if(!h_rsst || !status){
 	RsstSetRtCode(rsst, RSS_RT_BAD_HANDLE_STATUS);
 	return;
 }

	StringFitIn( obj, RsstKeyGet0(h_rsst, "RegName"), sizeof(obj) );
#ifdef WIN32
 if( !memcmp("HKEY_", obj, 5) ){
	if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegOpenKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else if( !strcmp("HKEY_CURRENT_CONFIG", obj) )
	rt= RegOpenKey( HKEY_CURRENT_CONFIG, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else if( !strcmp("HKEY_CURRENT_USER", obj) )
	rt= RegOpenKey( HKEY_CURRENT_USER, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else if( !strcmp("HKEY_LOCAL_MACHINE", obj) )
	rt= RegOpenKey( HKEY_LOCAL_MACHINE, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else if( !strcmp("HKEY_USERS", obj) )
	rt= RegOpenKey( HKEY_USERS, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else
	rt=RSS_RT_INVALID_REG_KEY;

  if(rt==ERROR_SUCCESS){
  	datalen= sizeof(s);
  	datatype=REG_SZ;
	rt= RegQueryValueEx( hkey, RsstKeyGet0(rsst, "SubName"), 0, &datatype, s, &datalen );
	RegCloseKey(hkey);  	
	if(rt==ERROR_SUCCESS)
	RsstKeyPut0( rsst, "Value", s );
	else
	RsstSetRtCode( rsst, RSS_RT_INVALID_REG_KEY );
  }
  else{
	RsstSetRtCode( rsst, RSS_RT_INVALID_REG_KEY );
  }
 }
 else{
	StringFitIn( s1, RsstKeyGet0(rsst, "SectionName"), sizeof(s1) );
 	GetPrivateProfileString(s1, RsstKeyGet0(rsst, "KeyName"), "", s, sizeof(s), obj );

	RsstKeyPut0( rsst, "Value", s );
 }
#else
	StringFitIn( s1, RsstKeyGet0(rsst, "SectionName"), sizeof(s1) );
	IniGetLine( s1, RsstKeyGet0(rsst, "KeyName"), "", s, sizeof(s), obj );

	RsstKeyPut0( rsst, "Value", s );
#endif
}
else if( !strcmp( act, "WriteKey" ) ){
	h_rsst= RsstGetHandle(rsst);
	status= RsstGetStatus(rsst);
 if(!h_rsst || !status){
 	RsstSetRtCode(rsst, RSS_RT_BAD_HANDLE_STATUS);
 	return;
 }

	StringFitIn( obj, RsstKeyGet0(h_rsst, "RegName"), sizeof(obj) );
#ifdef WIN32
 if( !memcmp("HKEY_", obj, 5) ){
	if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegCreateKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegCreateKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegCreateKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegCreateKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegCreateKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName"), &hkey );
	else
	rt=RSS_RT_INVALID_REG_KEY;

  if(rt==ERROR_SUCCESS){
  	datalen= sizeof(s);
	rt= RegSetValueEx( hkey, RsstKeyGet0(rsst, "SubName"), 0, REG_SZ, s, datalen );
	RegCloseKey(hkey);  	
	if(rt!=ERROR_SUCCESS)
	RsstSetRtCode( rsst, RSS_RT_INVALID_REG_KEY );
  }
  else{
	RsstSetRtCode( rsst, RSS_RT_INVALID_REG_KEY );
  }
 }
 else{
	StringFitIn( s1, RsstKeyGet0(rsst, "SectionName"), sizeof(s1) );
	StringFitIn( s, RsstKeyGet0(rsst, "Value"), sizeof(s) );
 	WritePrivateProfileString( s1, RsstKeyGet0(rsst, "KeyName"), s, obj );
 }
#else
	StringFitIn( s1, RsstKeyGet0(rsst, "SectionName"), sizeof(s1) );
	StringFitIn( s, RsstKeyGet0(rsst, "Value"), sizeof(s) );
	IniSetLine( s1, RsstKeyGet0(rsst, "KeyName"), s, obj );
#endif
}
else if( !strcmp( act, "DeleteKey" ) ){
	h_rsst= RsstGetHandle(rsst);
	status= RsstGetStatus(rsst);
 if(!h_rsst || !status){
 	RsstSetRtCode(rsst, RSS_RT_BAD_HANDLE_STATUS);
 	return;
 }

	StringFitIn( obj, RsstKeyGet0(h_rsst, "RegName"), sizeof(obj) );
#ifdef WIN32
 if( !memcmp("HKEY_", obj, 5) ){
	if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegDeleteKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName") );
	else if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegDeleteKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName") );
	else if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegDeleteKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName") );
	else if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegDeleteKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName") );
	else if( !strcmp("HKEY_CLASSES_ROOT", obj) )
	rt= RegDeleteKey( HKEY_CLASSES_ROOT, RsstKeyGet0(rsst, "KeyName") );
	else
	rt=RSS_RT_INVALID_REG_KEY;

	if(rt!=ERROR_SUCCESS)
	RsstSetRtCode( rsst, RSS_RT_INVALID_REG_KEY );
 }
 else{
	StringFitIn( s1, RsstKeyGet0(rsst, "SectionName"), sizeof(s1) );
 	WritePrivateProfileString(s1, RsstKeyGet0(rsst, "KeyName"), "", obj );
 }
#else
	StringFitIn( s1, RsstKeyGet0(rsst, "SectionName"), sizeof(s1) );
	IniSetLine( s1, RsstKeyGet0(rsst, "KeyName"), "", obj );
#endif
}
else{
	RsstSetRtCode( rsst, RSS_RT_UNKNOWN_ACTION );
}

}
